@@ -28,22 +28,15 @@ How to use Panel widgets in 3 simple steps
 For more detail see the Reference Gallery guide.
 https://panel.holoviz.org/reference/chat/ChatInterface.html
 """
-import importlib as _importlib
+
+from typing import TYPE_CHECKING
 
 from .feed import ChatFeed  # noqa
 from .icon import ChatReactionIcons  # noqa
 from .input import ChatAreaInput  # noqa
 from .interface import ChatInterface  # noqa
 from .message import ChatMessage  # noqa
-
-
-def __getattr__(name):
-    """
-    Lazily import langchain module when accessed.
-    """
-    if name == 'langchain':
-        return _importlib.import_module('panel.chat.langchain')
-    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
+from .step import ChatStep  # noqa
 
 __all__ = (
     "ChatAreaInput",
@@ -51,5 +44,20 @@ __all__ = (
     "ChatInterface",
     "ChatMessage",
     "ChatReactionIcons",
+    "ChatStep",
     "langchain",
 )
+
+def __getattr__(name):
+    """
+    Lazily import langchain module when accessed.
+    """
+    if name == "langchain":
+        from . import langchain
+        return langchain
+    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
+
+__dir__ = lambda: list(__all__)
+
+if TYPE_CHECKING:
+    from . import langchain
