@@ -9,6 +9,7 @@ import base64
 import logging
 import os
 import pathlib
+import sys
 
 from glob import glob
 from types import ModuleType
@@ -86,7 +87,7 @@ class AdminApplicationContext(ApplicationContext):
 
 class Serve(_BkServe):
 
-    args = _BkServe.args + (
+    args = tuple((arg, arg_obj) for arg, arg_obj in _BkServe.args if arg != '--dev') + (
         ('--static-dirs', dict(
             metavar="KEY=VALUE",
             nargs='+',
@@ -231,6 +232,10 @@ class Serve(_BkServe):
             type    = str,
             help    = "The profiler to use by default, e.g. pyinstrument, snakeviz or memray."
         )),
+        ('--dev', dict(
+            action  = 'store_true',
+            help    = "Whether to enable dev mode. Equivalent to --autoreload."
+        )),
         ('--autoreload', dict(
             action  = 'store_true',
             help    = "Whether to autoreload source when script changes."
@@ -280,14 +285,16 @@ class Serve(_BkServe):
                 applications['/'] = applications[f'/{index}']
         return super().customize_applications(args, applications)
 
-    def warm_applications(self, applications, reuse_sessions, error=True):
+    def warm_applications(self, applications, reuse_sessions, error=True, initialize_session=True):
         from ..io.session import generate_session
         for path, app in applications.items():
             try:
-                session = generate_session(app)
+                session = generate_session(app, initialize=initialize_session)
             except Exception as e:
                 if error:
                     raise e
+                else:
+                    continue
             with set_curdoc(session.document):
                 if config.session_key_func:
                     reuse_sessions = False
@@ -348,7 +355,7 @@ class Serve(_BkServe):
             pattern = REST_PROVIDERS[args.rest_provider](files, args.rest_endpoint)
             patterns.extend(pattern)
         elif args.rest_provider is not None:
-            raise ValueError("rest-provider %r not recognized." % args.rest_provider)
+            raise ValueError(f"rest-provider {args.rest_provider!r} not recognized.")
 
         config.global_loading_spinner = args.global_loading_spinner
         config.reuse_sessions = args.reuse_sessions
@@ -358,32 +365,31 @@ class Serve(_BkServe):
                 watch(f)
 
         if args.setup:
-            setup_path = args.setup
-            with open(setup_path) as f:
-                setup_source = f.read()
-            nodes = ast.parse(setup_source, os.fspath(setup_path))
-            code = compile(nodes, filename=setup_path, mode='exec', dont_inherit=True)
             module_name = 'panel_setup_module'
             module = ModuleType(module_name)
-            module.__dict__['__file__'] = fullpath(setup_path)
-            exec(code, module.__dict__)
+            module.__dict__['__file__'] = fullpath(args.setup)
             state._setup_module = module
 
-        if args.warm or args.autoreload:
+        if args.warm or config.autoreload:
             argvs = {f: args.args for f in files}
             applications = build_single_handler_applications(files, argvs)
-            if args.autoreload:
+            initialize_session = not (args.num_procs and sys.version_info < (3, 12))
+            if config.autoreload:
                 with record_modules(list(applications.values())):
                     self.warm_applications(
-                        applications, args.reuse_sessions, error=False
+                        applications, args.reuse_sessions, error=False, initialize_session=initialize_session
                     )
             else:
-                self.warm_applications(applications, args.reuse_sessions)
+                self.warm_applications(applications, args.reuse_sessions, initialize_session=initialize_session)
+
+        # Disable Tornado's autoreload
+        if args.dev:
+            del server_kwargs['autoreload']
 
         if args.liveness:
             argvs = {f: args.args for f in files}
             applications = build_single_handler_applications(files, argvs)
-            patterns += [(r"/%s" % args.liveness_endpoint, LivenessHandler, dict(applications=applications))]
+            patterns += [(rf"/{args.liveness_endpoint}", LivenessHandler, dict(applications=applications))]
 
         config.profiler = args.profiler
         if args.admin:
@@ -645,8 +651,9 @@ class Serve(_BkServe):
     def invoke(self, args: argparse.Namespace):
         # Autoreload must be enabled before the application(s) are executed
         # to avoid erroring out
-        config.autoreload = args.autoreload
+        config.autoreload = args.autoreload or bool(args.dev)
         # Empty layout are valid and the Bokeh warning is silenced as usually
         # not relevant to Panel users.
         silence(EMPTY_LAYOUT, True)
+        args.dev = None
         super().invoke(args)
