@@ -3,6 +3,7 @@ Utilities for creating bokeh Server instances.
 """
 from __future__ import annotations
 
+import ast
 import asyncio
 import datetime as dt
 import importlib
@@ -12,15 +13,13 @@ import os
 import pathlib
 import signal
 import sys
-import threading
 import uuid
 
 from contextlib import contextmanager
 from functools import partial, wraps
 from html import escape
-from types import FunctionType, MethodType
 from typing import (
-    TYPE_CHECKING, Any, Callable, Mapping, Optional, Union,
+    TYPE_CHECKING, Any, Callable, Mapping, Optional,
 )
 from urllib.parse import urljoin, urlparse
 
@@ -29,8 +28,6 @@ import param
 import tornado
 
 # Bokeh imports
-from bokeh.application import Application as BkApplication
-from bokeh.application.handlers.function import FunctionHandler
 from bokeh.core.json_encoder import serialize_json
 from bokeh.core.templates import AUTOLOAD_JS, FILE, MACROS
 from bokeh.core.validation import silence
@@ -39,7 +36,6 @@ from bokeh.embed.bundle import Script
 from bokeh.embed.elements import script_for_render_items
 from bokeh.embed.util import RenderItem
 from bokeh.embed.wrappers import wrap_in_script_tag
-from bokeh.models import CustomJS
 from bokeh.server.server import Server as BokehServer
 from bokeh.server.urls import per_app_patterns, toplevel_patterns
 from bokeh.server.views.autoload_js_handler import (
@@ -63,15 +59,13 @@ from tornado.wsgi import WSGIContainer
 from ..config import config
 from ..util import edit_readonly, fullpath
 from ..util.warnings import warn
-from .application import Application, build_single_handler_application
+from .application import build_applications
 from .document import (  # noqa
     _cleanup_doc, init_doc, unlocked, with_lock,
 )
 from .liveness import LivenessHandler
 from .loading import LOADING_INDICATOR_CSS_CLASS
-from .logging import (
-    LOG_SESSION_CREATED, LOG_SESSION_DESTROYED, LOG_SESSION_LAUNCHING,
-)
+from .logging import LOG_SESSION_CREATED
 from .reload import record_modules
 from .resources import (
     BASE_TEMPLATE, CDN_DIST, COMPONENT_PATH, ERROR_TEMPLATE, LOCAL_DIST,
@@ -79,23 +73,21 @@ from .resources import (
 )
 from .session import generate_session
 from .state import set_curdoc, state
+from .threads import StoppableThread
 
 logger = logging.getLogger(__name__)
 
 if TYPE_CHECKING:
     from bokeh.bundle import Bundle
     from bokeh.core.types import ID
-    from bokeh.document.document import DocJson, Document
+    from bokeh.document.document import DocJson
     from bokeh.server.contexts import BokehSessionContext
     from bokeh.server.session import ServerSession
     from jinja2 import Template
 
-    from ..template.base import BaseTemplate
-    from ..viewable import Viewable, Viewer
+    from .application import TViewableFuncOrPath
     from .location import Location
 
-    TViewable = Union[Viewable, Viewer, BaseTemplate]
-    TViewableFuncOrPath = Union[TViewable, Callable[[], TViewable], os.PathLike, str]
 
 #---------------------------------------------------------------------
 # Private API
@@ -115,38 +107,6 @@ def _server_url(url: str, port: int) -> str:
     else:
         return 'http://%s:%d%s' % (url.split(':')[0], port, "/")
 
-def _eval_panel(
-    panel: TViewableFuncOrPath, server_id: str, title: str,
-    location: bool | Location, admin: bool, doc: Document
-):
-    from ..pane import panel as as_panel
-    from ..template import BaseTemplate
-
-    if config.global_loading_spinner:
-        doc.js_on_event(
-            'document_ready', CustomJS(code=f"""
-            const body = document.getElementsByTagName('body')[0]
-            body.classList.remove({LOADING_INDICATOR_CSS_CLASS!r}, {config.loading_spinner!r})
-            """)
-        )
-
-    doc.on_event('document_ready', partial(state._schedule_on_load, doc))
-
-    # Set up instrumentation for logging sessions
-    logger.info(LOG_SESSION_LAUNCHING, id(doc))
-    def _log_session_destroyed(session_context):
-        logger.info(LOG_SESSION_DESTROYED, id(doc))
-    doc.on_session_destroyed(_log_session_destroyed)
-
-    with set_curdoc(doc):
-        if isinstance(panel, (FunctionType, MethodType)):
-            panel = panel()
-        if isinstance(panel, BaseTemplate):
-            doc = panel._modify_doc(server_id, title, doc, location)
-        else:
-            doc = as_panel(panel)._modify_doc(server_id, title, doc, location)
-        return doc
-
 def async_execute(func: Callable[..., None]) -> None:
     """
     Wrap async event loop scheduling to ensure that with_lock flag
@@ -335,10 +295,20 @@ class Server(BokehServer):
         if state._admin_context:
             state._admin_context._loop = self._loop
 
+    def setup_file(self):
+        setup_path = state._setup_module.__dict__['__file__']
+        with open(setup_path) as f:
+            setup_source = f.read()
+        nodes = ast.parse(setup_source, os.fspath(setup_path))
+        code = compile(nodes, filename=setup_path, mode='exec', dont_inherit=True)
+        exec(code, state._setup_module.__dict__)
+
     def start(self) -> None:
         super().start()
         if state._admin_context:
             self._loop.add_callback(state._admin_context.run_load_hook)
+        if state._setup_module:
+            self._loop.add_callback(self.setup_file)
         if config.autoreload:
             from .reload import setup_autoreload_watcher
             self._autoreload_stop_event = stop_event = asyncio.Event()
@@ -349,9 +319,15 @@ class Server(BokehServer):
             # For the stop event to be processed we have to restart
             # the IOLoop briefly, ensuring an orderly cleanup
             async def stop_autoreload():
+                for event in state._watch_events:
+                    event.set()
+                state._watch_events = []
                 self._autoreload_stop_event.set()
                 await self._autoreload_task
-            self._loop.asyncio_loop.run_until_complete(stop_autoreload())
+            try:
+                self._loop.asyncio_loop.run_until_complete(stop_autoreload())
+            except RuntimeError:
+                pass # Ignore if the event loop is still running
         super().stop(wait=wait)
         if state._admin_context:
             state._admin_context.run_unload_hook()
@@ -800,9 +776,6 @@ def serve(
     kwargs: dict
       Additional keyword arguments to pass to Server instance
     """
-    # Empty layout are valid and the Bokeh warning is silenced as usually
-    # not relevant to Panel users.
-    silence(EMPTY_LAYOUT, True)
     kwargs = dict(kwargs, **dict(
         port=port, address=address, websocket_origin=websocket_origin,
         loop=loop, show=show, start=start, title=title, verbose=verbose,
@@ -857,9 +830,9 @@ def get_static_routes(static_dirs):
                              "this is reserved for internal use.")
         path = fullpath(path)
         if not os.path.isdir(path):
-            raise ValueError("Cannot serve non-existent path %s" % path)
+            raise ValueError(f"Cannot serve non-existent path {path}")
         patterns.append(
-            (r"%s/(.*)" % slug, StaticFileHandler, {"path": path})
+            (rf"{slug}/(.*)", StaticFileHandler, {"path": path})
         )
     patterns.append((
         f'/{COMPONENT_PATH}(.*)', ComponentResourceHandler, {}
@@ -874,7 +847,7 @@ def get_server(
     loop: Optional[IOLoop] = None,
     show: bool = False,
     start: bool = False,
-    title: bool = None,
+    title: str | dict[str, str] | None = None,
     verbose: bool = False,
     location: bool | Location = True,
     admin: bool = False,
@@ -1001,60 +974,34 @@ def get_server(
     from ..config import config
     from .rest import REST_PROVIDERS
 
+    silence(EMPTY_LAYOUT, True)
     server_id = kwargs.pop('server_id', uuid.uuid4().hex)
-    kwargs['extra_patterns'] = extra_patterns = kwargs.get('extra_patterns', [])
-    if isinstance(panel, dict):
-        apps = {}
-        for slug, app in panel.items():
-            if slug.endswith('/') and not slug == '/':
-                raise ValueError(f"Invalid URL: trailing slash '/' used for {slug!r} not supported.")
-            if isinstance(title, dict):
-                try:
-                    title_ = title[slug]
-                except KeyError:
-                    raise KeyError(
-                        "Keys of the title dictionary and of the apps "
-                        f"dictionary must match. No {slug} key found in the "
-                        "title dictionary.") from None
-            else:
-                title_ = title
-            slug = slug if slug.startswith('/') else '/'+slug
-            if 'flask' in sys.modules:
-                from flask import Flask
-                if isinstance(app, Flask):
-                    wsgi = WSGIContainer(app)
-                    if slug == '/':
-                        raise ValueError('Flask apps must be served on a subpath.')
-                    if not slug.endswith('/'):
-                        slug += '/'
-                    extra_patterns.append(('^'+slug+'.*', ProxyFallbackHandler,
-                                           dict(fallback=wsgi, proxy=slug)))
-                    continue
-            if isinstance(app, pathlib.Path):
-                app = str(app) # enables serving apps from Paths
-            if (isinstance(app, str) and app.endswith(('.py', '.ipynb', '.md'))
-                and os.path.isfile(app)):
-                apps[slug] = app = build_single_handler_application(app)
-                app._admin = admin
-            elif isinstance(app, BkApplication):
-                apps[slug] = app
-            else:
-                handler = FunctionHandler(partial(_eval_panel, app, server_id, title_, location, admin))
-                apps[slug] = Application(handler, admin=admin)
-    else:
-        if isinstance(panel, pathlib.Path):
-            panel = str(panel) # enables serving apps from Paths
-        if isinstance(panel, BkApplication):
-            panel = {'/': app}
-        elif (isinstance(panel, str) and panel.endswith(('.py', '.ipynb', '.md'))
-            and os.path.isfile(panel)):
-            apps = {'/': build_single_handler_application(panel)}
-        else:
-            handler = FunctionHandler(partial(_eval_panel, panel, server_id, title, location, admin))
-            apps = {'/': Application(handler, admin=admin)}
+    kwargs['extra_patterns'] = extra_patterns = list(kwargs.get('extra_patterns', []))
+
+    def flask_handler(slug, app):
+        if 'flask' not in sys.modules:
+            return
+        from flask import Flask
+        if not isinstance(app, Flask):
+            return
+        wsgi = WSGIContainer(app)
+        if slug == '/':
+            raise ValueError('Flask apps must be served on a subpath.')
+        if not slug.endswith('/'):
+            slug += '/'
+        extra_patterns.append((
+            f'^{slug}.*', ProxyFallbackHandler, dict(fallback=wsgi, proxy=slug)
+        ))
+        return True
+
+    apps = build_applications(
+        panel, title=title, location=location, admin=admin, custom_handlers=(flask_handler,)
+    )
 
     if warm or config.autoreload:
-        for app in apps.values():
+        for endpoint, app in apps.items():
+            if endpoint == '/admin':
+                continue
             if config.autoreload:
                 with record_modules(list(apps.values())):
                     session = generate_session(app)
@@ -1064,16 +1011,6 @@ def get_server(
                 state._on_load(None)
             _cleanup_doc(session.document, destroy=True)
 
-    if admin:
-        if '/admin' in apps:
-            raise ValueError(
-                'Cannot enable admin panel because another app is being served '
-                'on the /admin endpoint'
-            )
-        from .admin import admin_panel
-        admin_handler = FunctionHandler(admin_panel)
-        apps['/admin'] = Application(admin_handler)
-
     extra_patterns += get_static_routes(static_dirs)
 
     if session_history is not None:
@@ -1085,7 +1022,7 @@ def get_server(
 
     if liveness:
         liveness_endpoint = 'liveness' if isinstance(liveness, bool) else liveness
-        extra_patterns += [(r"/%s" % liveness_endpoint, LivenessHandler, dict(applications=apps))]
+        extra_patterns += [(rf"/{liveness_endpoint}", LivenessHandler, dict(applications=apps))]
 
     opts = dict(kwargs)
     if loop:
@@ -1160,7 +1097,7 @@ def get_server(
         server.io_loop.add_callback(show_callback)
 
     def sig_exit(*args, **kwargs):
-        server.io_loop.add_callback_from_signal(do_stop)
+        server.io_loop.asyncio_loop.call_soon_threadsafe(do_stop)
 
     def do_stop(*args, **kwargs):
         server.io_loop.stop()
@@ -1182,35 +1119,3 @@ def get_server(
                 "process invoking the panel.io.server.serve."
             )
     return server
-
-
-class StoppableThread(threading.Thread):
-    """Thread class with a stop() method."""
-
-    def __init__(self, io_loop: IOLoop, **kwargs):
-        super().__init__(**kwargs)
-        self.io_loop = io_loop
-
-    def run(self) -> None:
-        if hasattr(self, '_target'):
-            target, args, kwargs = self._target, self._args, self._kwargs # type: ignore
-        else:
-            target, args, kwargs = self._Thread__target, self._Thread__args, self._Thread__kwargs # type: ignore
-        if not target:
-            return
-        bokeh_server = None
-        try:
-            bokeh_server = target(*args, **kwargs)
-        finally:
-            if isinstance(bokeh_server, Server):
-                try:
-                    bokeh_server.stop()
-                except Exception:
-                    pass
-            if hasattr(self, '_target'):
-                del self._target, self._args, self._kwargs # type: ignore
-            else:
-                del self._Thread__target, self._Thread__args, self._Thread__kwargs # type: ignore
-
-    def stop(self) -> None:
-        self.io_loop.add_callback(self.io_loop.stop)
