@@ -148,7 +148,7 @@ function group_data(records: any[], columns: any[], indexes: string[], aggregato
 }
 
 const timestampSorter = function(a: any, b: any, _aRow: any, _bRow: any, _column: any, _dir: any, _params: any) {
-  // Bokeh serializes datetime objects as UNIX timestamps.
+  // Bokeh/Panel serializes datetime objects as UNIX timestamps (in milliseconds).
 
   //a, b - the two values being compared
   //aRow, bRow - the row components for the values being compared (useful if you need to access additional fields in the row data for the sort)
@@ -164,14 +164,12 @@ const timestampSorter = function(a: any, b: any, _aRow: any, _bRow: any, _column
 
   const opts = {zone: new (window as any).luxon.IANAZone("UTC")}
 
-  // NaN values are serialized to -9223372036854776 by Bokeh
-
-  if (String(a) == "-9223372036854776") {
+  if (Number.isNaN(a)) {
     a = (window as any).luxon.DateTime.fromISO("invalid")
   } else {
     a = (window as any).luxon.DateTime.fromMillis(a, opts)
   }
-  if (String(b) == "-9223372036854776") {
+  if (Number.isNaN(b)) {
     b = (window as any).luxon.DateTime.fromISO("invalid")
   } else {
     b = (window as any).luxon.DateTime.fromMillis(b, opts)
@@ -304,6 +302,44 @@ const datetimeEditor = function(cell: any, onRendered: any, success: any, cancel
   return input
 }
 
+const nestedEditor = function(cell: any, editorParams: any) {
+  //cell - the cell component for the editable cell
+
+  const row = cell.getRow().getData()
+  let values = editorParams.options
+  for (const i of editorParams.lookup_order) {
+    values = row[i] in values ? values[row[i]] : []
+    if (Array.isArray(values)) {
+      break
+    }
+  }
+  return values ? values : []
+}
+
+function find_column(group: any, field: string): any {
+  if (group.columns != null) {
+    for (const col of group.columns) {
+      const found = find_column(col, field)
+      if (found) {
+        return found
+      }
+    }
+  } else {
+    return group.field === field ? group : null
+  }
+}
+
+function clone_column(group: any): any {
+  if (group.columns == null) {
+    return {...group}
+  }
+  const group_columns = []
+  for (const col of group.columns) {
+    group_columns.push(clone_column(col))
+  }
+  return {...group, columns: group_columns}
+}
+
 export class DataTabulatorView extends HTMLBoxView {
   declare model: DataTabulator
 
@@ -314,13 +350,16 @@ export class DataTabulatorView extends HTMLBoxView {
   _updating_page: boolean = false
   _updating_sort: boolean = false
   _selection_updating: boolean = false
+  _last_selected_row: any = null
   _initializing: boolean
   _lastVerticalScrollbarTopPosition: number = 0
   _lastHorizontalScrollbarLeftPosition: number = 0
   _applied_styles: boolean = false
   _building: boolean = false
   _debounced_redraw: any = null
-  _restore_scroll: boolean = false
+  _restore_scroll: boolean | "horizontal" | "vertical" = false
+  _updating_scroll: boolean = false
+  _is_scrolling: boolean = false
 
   override connect_signals(): void {
     super.connect_signals()
@@ -365,7 +404,7 @@ export class DataTabulatorView extends HTMLBoxView {
       for (const row of this.tabulator.rowManager.getRows()) {
         if (row.cells.length > 0) {
           const index = row.data._index
-          const icon = this.model.expanded.indexOf(index) < 0 ? "►" : "▼"
+          const icon = this.model.expanded.includes(index) ? "▼" : "►"
           row.cells[1].element.innerText = icon
         }
       }
@@ -397,18 +436,23 @@ export class DataTabulatorView extends HTMLBoxView {
       if (this.tabulator === undefined) {
         return
       }
+      this._restore_scroll = "horizontal"
       this._selection_updating = true
+      this._updating_scroll = true
       this.setData()
+      this._updating_scroll = false
       this._selection_updating = false
       this.postUpdate()
     })
     this.connect(this.model.source.streaming, () => this.addData())
     this.connect(this.model.source.patching, () => {
       const inds = this.model.source.selected.indices
+      this._updating_scroll = true
       this.updateOrAddData()
-      this.record_scroll()
+      this._updating_scroll = false
       // Restore indices since updating data may have reset checkbox column
       this.model.source.selected.indices = inds
+      this.restore_scroll()
     })
     this.connect(this.model.source.selected.change, () => this.setSelection())
     this.connect(this.model.source.selected.properties.indices.change, () => this.setSelection())
@@ -471,7 +515,9 @@ export class DataTabulatorView extends HTMLBoxView {
 
   override after_resize(): void {
     super.after_resize()
-    this._debounced_redraw()
+    if (!this._is_scrolling) {
+      this._debounced_redraw()
+    }
   }
 
   _resize_redraw(): void {
@@ -500,8 +546,8 @@ export class DataTabulatorView extends HTMLBoxView {
     }
     super.render()
     this._initializing = true
-    const container = div({style: "display: contents;"})
-    const el = div({style: "width: 100%; height: 100%; visibility: hidden;"})
+    const container = div({style: {display: "contents"}})
+    const el = div({style: {width: "100%", height: "100%", visibility: "hidden"}})
     this.container = el
     this.setCSSClasses(el)
     container.appendChild(el)
@@ -540,7 +586,7 @@ export class DataTabulatorView extends HTMLBoxView {
     this.tabulator.on("tableBuilt", () => this.tableBuilt())
 
     // Rendering callbacks
-    this.tabulator.on("selectableCheck", (row: any) => {
+    this.tabulator.on("selectableRowsCheck", (row: any) => {
       const selectable = this.model.selectable_rows
       return (selectable == null) || selectable.includes(row._row.data._index)
     })
@@ -548,12 +594,8 @@ export class DataTabulatorView extends HTMLBoxView {
       return `${cell.getColumn().getField()}: ${cell.getValue()}`
     })
     this.tabulator.on("scrollVertical", debounce(() => {
-      this.record_scroll()
       this.setStyles()
     }, 50, false))
-    this.tabulator.on("scrollHorizontal", debounce(() => {
-      this.record_scroll()
-    }, 50, false))
 
     // Sync state with model
     this.tabulator.on("rowSelectionChanged", (data: any, rows: any, selected: any, deselected: any) => this.rowSelectionChanged(data, rows, selected, deselected))
@@ -571,8 +613,10 @@ export class DataTabulatorView extends HTMLBoxView {
       if (rows.length === 0) {
         this.tabulator.rowManager.renderEmptyScroll()
       }
-      // Ensure that after filtering the page is updated
-      this.updatePage(this.tabulator.getPage())
+      if (this.model.pagination != null) {
+        // Ensure that after filtering the page is updated
+        this.updatePage(this.tabulator.getPage())
+      }
     })
     this.tabulator.on("pageLoaded", (pageno: number) => {
       this.updatePage(pageno)
@@ -604,7 +648,45 @@ export class DataTabulatorView extends HTMLBoxView {
     this.renderChildren()
     this.setStyles()
 
+    // Track scrolling position and active scroll
+    const holder = this.shadow_el.querySelector(".tabulator-tableholder")
+    let scroll_timeout: ReturnType<typeof setTimeout> | undefined
+    if (holder) {
+      holder.addEventListener("scroll", () => {
+        this.record_scroll()
+        this._is_scrolling = true
+        clearTimeout(scroll_timeout)
+        scroll_timeout = setTimeout(() => {
+          this._is_scrolling = false
+        }, 200)
+      })
+    }
+
     if (this.model.pagination) {
+      if (this.model.page_size == null) {
+        const table = this.shadow_el.querySelector(".tabulator-table")
+        if (table != null && holder != null) {
+          const table_height = holder.clientHeight
+          let height = 0
+          let page_size = null
+          const heights = []
+          for (let i = 0; i<table.children.length; i++) {
+            const row_height = table.children[i].clientHeight
+            heights.push(row_height)
+            height += row_height
+            if (height > table_height) {
+              page_size = i
+              break
+            }
+          }
+          if (height < table_height) {
+            page_size = table.children.length
+            const remaining = table_height - height
+            page_size += Math.floor(remaining / Math.min(...heights))
+          }
+          this.model.page_size = Math.max(page_size || 1, 1)
+        }
+      }
       this.setMaxPage()
       this.tabulator.setPage(this.model.page)
     }
@@ -655,22 +737,21 @@ export class DataTabulatorView extends HTMLBoxView {
 
   getConfiguration(): any {
     // Only use selectable mode if explicitly requested otherwise manually handle selections
-    const selectable = this.model.select_mode === "toggle" ? true : NaN
+    const selectableRows = this.model.select_mode === "toggle" ? true : NaN
     const configuration = {
       ...this.model.configuration,
       index: "_index",
       nestedFieldSeparator: false,
       movableColumns: false,
-      selectable,
+      selectableRows,
       columns: this.getColumns(),
       initialSort: this.sorters,
       layout: this.getLayout(),
       pagination: this.model.pagination != null,
       paginationMode: this.model.pagination,
-      paginationSize: this.model.page_size,
+      paginationSize: this.model.page_size || 20,
       paginationInitialPage: 1,
       groupBy: this.groupBy,
-      rowFormatter: (row: any) => this._render_row(row),
       frozenRows: (row: any) => {
         return (this.model.frozen_rows.length > 0) ? this.model.frozen_rows.includes(row._row.data._index) : false
       },
@@ -695,10 +776,17 @@ export class DataTabulatorView extends HTMLBoxView {
     }
   }
 
+  get_child(idx: number): LayoutDOM | null {
+    if (this.model.children instanceof Map) {
+      return this.model.children.get(idx) || null
+    }
+    return null
+  }
+
   override get child_models(): LayoutDOM[] {
     const children: LayoutDOM[] = []
     for (const idx of this.model.expanded) {
-      const child = this.model.children.get(idx)
+      const child = this.get_child(idx)
       if (child != null) {
         children.push(child)
       }
@@ -711,16 +799,21 @@ export class DataTabulatorView extends HTMLBoxView {
       const new_children = await this.build_child_views()
       resolve(new_children)
     }).then((new_children) => {
-      for (const r of this.model.expanded) {
-        const row = this.tabulator.getRow(r)
+      const rows = this.tabulator.getRows()
+      const lookup = new Map()
+      for (const row of rows) {
         const index = row._row?.data._index
-        if (this.model.children.get(index) == null) {
-          continue
+        if (index != null) {
+          lookup.set(index, row)
         }
-        const model = this.model.children.get(index)
+      }
+      for (const index of this.model.expanded) {
+        const model = this.get_child(index)
+        const row = lookup.get(index)
         const view = model == null ? null : this._child_views.get(model)
-        if ((view != null) && (new_children as UIElementView[]).includes(view)) {
-          this._render_row(row, false)
+        if (view != null) {
+          const render = (new_children as UIElementView[]).includes(view)
+          this._render_row(row, false, render)
         }
       }
       this._update_children()
@@ -731,12 +824,12 @@ export class DataTabulatorView extends HTMLBoxView {
     })
   }
 
-  _render_row(row: any, resize: boolean = true): void {
+  _render_row(row: any, resize: boolean = true, render: boolean = true): void {
     const index = row._row?.data._index
-    if (!this.model.expanded.includes(index) || this.model.children.get(index) == null) {
+    if (!this.model.expanded.includes(index)) {
       return
     }
-    const model = this.model.children.get(index)
+    const model = this.get_child(index)
     const view = model == null ? null : this._child_views.get(model)
     if (view == null) {
       return
@@ -745,10 +838,16 @@ export class DataTabulatorView extends HTMLBoxView {
     const style = getComputedStyle(this.tabulator.element.children[1].children[0])
     const bg = style.backgroundColor
     const neg_margin = rowEl.style.paddingLeft ? `-${rowEl.style.paddingLeft}` : "0"
-    const viewEl = div({style: `background-color: ${bg}; margin-left:${neg_margin}; max-width: 100%; overflow-x: hidden;`})
+    const prev_child = rowEl.children[rowEl.children.length-1]
+    let viewEl
+    if (prev_child != null && prev_child.className == "row-content") {
+      viewEl = prev_child
+    } else {
+      viewEl = div({class: "row-content", style: {background_color: bg, margin_left: neg_margin, max_width: "100%", overflow_x: "hidden"}})
+    }
     viewEl.appendChild(view.el)
     rowEl.appendChild(viewEl)
-    if (!view.has_finished()) {
+    if (!view.has_finished() && render) {
       view.render()
       view.after_render()
     }
@@ -762,18 +861,18 @@ export class DataTabulatorView extends HTMLBoxView {
   _expand_render(cell: any): string {
     const index = cell._cell.row.data._index
     const icon = this.model.expanded.indexOf(index) < 0 ? "►" : "▼"
-    return `<i>${icon}</i>`
+    return icon
   }
 
   _update_expand(cell: any): void {
     const index = cell._cell.row.data._index
     const expanded = [...this.model.expanded]
-    const exp_index = expanded.indexOf(index)
-    if (exp_index < 0) {
+    if (!expanded.includes(index)) {
       expanded.push(index)
     } else {
+      const exp_index = expanded.indexOf(index)
       const removed = expanded.splice(exp_index, 1)[0]
-      const model = this.model.children.get(removed)
+      const model = this.get_child(removed)
       if (model != null) {
         const view = this._child_views.get(model)
         if (view !== undefined && view.el != null) {
@@ -782,12 +881,12 @@ export class DataTabulatorView extends HTMLBoxView {
       }
     }
     this.model.expanded = expanded
-    if (expanded.indexOf(index) < 0) {
+    if (!expanded.includes(index)) {
       return
     }
     let ready = true
     for (const idx of this.model.expanded) {
-      if (this.model.children.get(idx) == null) {
+      if (this.get_child(idx) == null) {
         ready = false
         break
       }
@@ -812,13 +911,8 @@ export class DataTabulatorView extends HTMLBoxView {
     columns.push({field: "_index", frozen: true, visible: false})
     if (config_columns != null) {
       for (const column of config_columns) {
-        if (column.columns != null) {
-          const group_columns = []
-          for (const col of column.columns) {
-            group_columns.push({...col})
-          }
-          columns.push({...column, columns: group_columns})
-        } else if (column.formatter === "expand") {
+        const new_column = clone_column(column)
+        if (column.formatter === "expand") {
           const expand = {
             hozAlign: "center",
             cellClick: (_: any, cell: any) => {
@@ -832,7 +926,6 @@ export class DataTabulatorView extends HTMLBoxView {
           }
           columns.push(expand)
         } else {
-          const new_column = {...column}
           if (new_column.formatter === "rowSelection") {
             new_column.cellClick = (_: any, cell: any) => {
               cell.getRow().toggleSelect()
@@ -846,18 +939,8 @@ export class DataTabulatorView extends HTMLBoxView {
       let tab_column: any = null
       if (config_columns != null) {
         for (const col of columns) {
-          if (col.columns != null) {
-            for (const c of col.columns) {
-              if (column.field === c.field) {
-                tab_column = c
-                break
-              }
-            }
-            if (tab_column != null) {
-              break
-            }
-          } else if (column.field === col.field) {
-            tab_column = col
+          tab_column = find_column(col, column.field)
+          if (tab_column != null) {
             break
           }
         }
@@ -907,6 +990,11 @@ export class DataTabulatorView extends HTMLBoxView {
           tab_column.editor = dateEditor
         } else if (tab_column.editor === "datetime") {
           tab_column.editor = datetimeEditor
+        } else if (tab_column.editor === "nested") {
+          tab_column.editorParams.valuesLookup = (cell: any) => {
+            return nestedEditor(cell, tab_column.editorParams)
+          }
+          tab_column.editor = "list"
         }
       } else if (ctype === "StringEditor") {
         if (editor.completions.length > 0) {
@@ -1004,16 +1092,15 @@ export class DataTabulatorView extends HTMLBoxView {
   }
 
   // Update table
-
-  setData(): void {
+  setData(): Promise<void> {
     if (this._initializing || this._building || !this.tabulator.initialized) {
-      return
+      return Promise.resolve(undefined)
     }
     const data = this.getData()
     if (this.model.pagination != null) {
-      this.tabulator.rowManager.setData(data, true, false)
+      return this.tabulator.rowManager.setData(data, true, false)
     } else {
-      this.tabulator.setData(data)
+      return this.tabulator.setData(data)
     }
   }
 
@@ -1021,9 +1108,20 @@ export class DataTabulatorView extends HTMLBoxView {
     const rows = this.tabulator.rowManager.getRows()
     const last_row = rows[rows.length-1]
     const start = ((last_row?.data._index) || 0)
-    this.setData()
-    if (this.model.follow && last_row) {
-      this.tabulator.scrollToRow(start, "top", false)
+    this._updating_page = true
+    const promise = this.setData()
+    if (this.model.follow) {
+      promise.then(() => {
+        if (this.model.pagination) {
+          this.tabulator.setPage(Math.ceil(this.tabulator.rowManager.getDataCount() / (this.model.page_size || 20)))
+        }
+        if (last_row) {
+          this.tabulator.scrollToRow(start, "top", false)
+        }
+        this._updating_page = false
+      })
+    } else {
+      this._updating_page = true
     }
   }
 
@@ -1031,7 +1129,9 @@ export class DataTabulatorView extends HTMLBoxView {
     this.setSelection()
     this.setStyles()
     if (this._restore_scroll) {
-      this.restore_scroll()
+      const vertical = this._restore_scroll === "horizontal" ? false : true
+      const horizontal = this._restore_scroll === "vertical" ? false : true
+      this.restore_scroll(horizontal, vertical)
       this._restore_scroll = false
     }
   }
@@ -1068,7 +1168,7 @@ export class DataTabulatorView extends HTMLBoxView {
   }
 
   updatePage(pageno: number): void {
-    if (this.model.pagination === "local" && this.model.page !== pageno) {
+    if (this.model.pagination === "local" && this.model.page !== pageno && !this._updating_page) {
       this._updating_page = true
       this.model.page = pageno
       this._updating_page = false
@@ -1179,18 +1279,30 @@ export class DataTabulatorView extends HTMLBoxView {
     this._selection_updating = false
   }
 
-  restore_scroll(): void {
-    const opts = {
-      top: this._lastVerticalScrollbarTopPosition,
-      left: this._lastHorizontalScrollbarLeftPosition,
-      behavior: "instant",
+  restore_scroll(horizontal: boolean=true, vertical: boolean=true): void {
+    if (!(horizontal || vertical)) {
+      return
     }
-    setTimeout(() => this.tabulator.rowManager.element.scrollTo(opts), 0)
+    const opts: ScrollToOptions = {behavior: "instant"}
+    if (vertical) {
+      opts.top = this._lastVerticalScrollbarTopPosition
+    }
+    if (horizontal) {
+      opts.left = this._lastHorizontalScrollbarLeftPosition
+    }
+    setTimeout(() => {
+      this._updating_scroll = true
+      this.tabulator.rowManager.element.scrollTo(opts)
+      this._updating_scroll = false
+    }, 0)
   }
 
   // Update model
 
   record_scroll() {
+    if (this._updating_scroll) {
+      return
+    }
     this._lastVerticalScrollbarTopPosition = this.tabulator.rowManager.element.scrollTop
     this._lastHorizontalScrollbarLeftPosition = this.tabulator.rowManager.element.scrollLeft
   }
@@ -1210,44 +1322,25 @@ export class DataTabulatorView extends HTMLBoxView {
     const selected = this.model.source.selected
     const index: number = row._row.data._index
 
-    if (this.model.pagination === "remote") {
-      const includes = this.model.source.selected.indices.indexOf(index) == -1
-      const flush = !(e.ctrlKey || e.metaKey || e.shiftKey)
-      if (e.shiftKey && selected.indices.length) {
-        const start = selected.indices[selected.indices.length-1]
-        if (index>start) {
-          for (let i = start; i<=index; i++) {
-            indices.push(i)
-          }
-        } else {
-          for (let i = start; i>=index; i--) {
-            indices.push(i)
-          }
-        }
-      } else {
-        indices.push(index)
-      }
-      this._selection_updating = true
-      this.model.trigger_event(new SelectionEvent(indices, includes, flush))
-      this._selection_updating = false
-      return
-    }
-
     if (e.ctrlKey || e.metaKey) {
-      indices = [...this.model.source.selected.indices]
-    } else if (e.shiftKey && selected.indices.length) {
-      const start = selected.indices[selected.indices.length-1]
-      if (index>start) {
-        for (let i = start; i<index; i++) {
-          indices.push(i)
-        }
-      } else {
-        for (let i = start; i>index; i--) {
-          indices.push(i)
-        }
+      indices = [...selected.indices]
+    } else if (e.shiftKey && this._last_selected_row) {
+      const rows = row._row.parent.getDisplayRows()
+      const start_idx = rows.indexOf(this._last_selected_row)
+      if (start_idx !== -1) {
+        const end_idx = rows.indexOf(row._row)
+        const reverse = start_idx > end_idx
+        const [start, end] = reverse ? [end_idx+1, start_idx+1] : [start_idx, end_idx]
+        indices = rows.slice(start, end).map((r: any) => r.data._index)
+        if (reverse) { indices = indices.reverse() }
       }
     }
-    if (indices.indexOf(index) < 0) {
+    const flush = !(e.ctrlKey || e.metaKey || e.shiftKey)
+    const includes = indices.includes(index)
+    const remote = this.model.pagination === "remote"
+
+    // Toggle the index on or off (if remote we let Python do the toggling)
+    if (!includes || remote) {
       indices.push(index)
     } else {
       indices.splice(indices.indexOf(index), 1)
@@ -1259,10 +1352,16 @@ export class DataTabulatorView extends HTMLBoxView {
       }
     }
     const filtered = this._filter_selected(indices)
-    this.tabulator.deselectRow()
-    this.tabulator.selectRow(filtered)
+    if (!remote) {
+      this.tabulator.deselectRow()
+      this.tabulator.selectRow(filtered)
+    }
+    this._last_selected_row = row._row
     this._selection_updating = true
-    selected.indices = filtered
+    if (!remote) {
+      selected.indices = filtered
+    }
+    this.model.trigger_event(new SelectionEvent(indices, !includes, flush))
     this._selection_updating = false
   }
 
@@ -1353,7 +1452,7 @@ export namespace DataTabulator {
     layout: p.Property<typeof TableLayout["__type__"]>
     max_page: p.Property<number>
     page: p.Property<number>
-    page_size: p.Property<number>
+    page_size: p.Property<number | null>
     pagination: p.Property<string | null>
     select_mode: p.Property<any>
     selectable_rows: p.Property<number[] | null>
@@ -1399,7 +1498,7 @@ export class DataTabulator extends HTMLBox {
       max_page:       [ Float,                   0 ],
       pagination:     [ Nullable(Str),      null ],
       page:           [ Float,                   0 ],
-      page_size:      [ Float,                   0 ],
+      page_size:      [ Nullable(Float),       null ],
       select_mode:    [ Any,                   true ],
       selectable_rows: [ Nullable(List(Float)), null ],
       source:         [ Ref(ColumnDataSource)       ],
