@@ -5,10 +5,12 @@ models rendered on the frontend.
 """
 from __future__ import annotations
 
+import asyncio
 import datetime as dt
 import difflib
 import inspect
 import logging
+import pathlib
 import re
 import sys
 import textwrap
@@ -20,6 +22,7 @@ from typing import (
     TYPE_CHECKING, Any, Callable, ClassVar, Mapping, Optional, Union,
 )
 
+import jinja2
 import numpy as np
 import param
 
@@ -46,18 +49,21 @@ from .models.reactive_html import (
 from .util import (
     HTML_SANITIZER, classproperty, edit_readonly, escape, updating,
 )
-from .viewable import Layoutable, Renderable, Viewable
+from .util.checks import import_available
+from .viewable import (
+    Child, Children, Layoutable, Renderable, Viewable,
+)
 
 if TYPE_CHECKING:
     import pandas as pd
 
     from bokeh.document import Document
     from bokeh.events import Event
-    from bokeh.model import Model
+    from bokeh.model import Model, ModelEvent
     from bokeh.models.sources import DataDict, Patches
     from pyviz_comms import Comm
 
-    from .layout.base import Panel
+    from .layout.base import Panel as BasePanel
     from .links import Callback, JSLinkTarget, Link
 
 log = logging.getLogger('panel.reactive')
@@ -132,6 +138,12 @@ class Syncable(Renderable):
         # A dictionary of bokeh property changes being processed
         self._changing = {}
 
+        # A dictionary of parameter changes being processed
+        self._in_process__events = {}
+
+        # Whether the component is watching the stylesheets
+        self._watching_stylesheets = False
+
         # Sets up watchers to process manual updates to models
         if self._manual_params:
             self._internal_callbacks.append(
@@ -279,7 +291,7 @@ class Syncable(Renderable):
         """
 
     def _update_manual(self, *events: param.parameterized.Event) -> None:
-        for ref, (model, parent) in self._models.items():
+        for ref, (model, parent) in self._models.copy().items():
             if ref not in state._views or ref in state._fake_roots:
                 continue
             viewable, root, doc, comm = state._views[ref]
@@ -295,21 +307,36 @@ class Syncable(Renderable):
                 else:
                     cb()
 
+    def _scheduled_update_model(
+        self, events: dict[str, param.parameterized.Event], msg: dict[str, Any],
+        root: Model, model: Model, doc: Document, comm: Optional[Comm],
+        curdoc_events: dict[str, Any]
+    ) -> None:
+        #
+        self._in_process__events[doc] = curdoc_events
+        try:
+            self._update_model(events, msg, root, model, doc, comm)
+        finally:
+            self._in_process__events.pop(doc, None)
+
     def _apply_update(
         self, events: dict[str, param.parameterized.Event], msg: dict[str, Any],
         model: Model, ref: str
-    ) -> None:
+    ) -> bool:
         if ref not in state._views or ref in state._fake_roots:
-            return
+            return True
         viewable, root, doc, comm = state._views[ref]
         if comm or not doc.session_context or state._unblocked(doc):
             with unlocked():
                 self._update_model(events, msg, root, model, doc, comm)
             if comm and 'embedded' not in root.tags:
                 push(doc, comm)
+            return True
         else:
-            cb = partial(self._update_model, events, msg, root, model, doc, comm)
+            curdoc_events = self._in_process__events.pop(doc, {})
+            cb = partial(self._scheduled_update_model, events, msg, root, model, doc, comm, curdoc_events)
             doc.add_next_tick_callback(cb)
+            return False
 
     def _update_model(
         self, events: dict[str, param.parameterized.Event], msg: dict[str, Any],
@@ -317,7 +344,20 @@ class Syncable(Renderable):
     ) -> None:
         ref = root.ref['id']
         self._changing[ref] = attrs = []
-        for attr, value in msg.items():
+        curdoc_events = self._in_process__events.get(doc, {})
+        for attr, value in msg.copy().items():
+            if attr in curdoc_events and value is curdoc_events[attr]:
+                # Do not apply change that originated directly from
+                # the frontend since this may cause boomerang if a
+                # new property value is already in-flight
+                del msg[attr]
+                continue
+            elif attr in self._events:
+                # Do not override a property value that was just changed
+                # on the frontend
+                del self._events[attr]
+                continue
+
             # Bokeh raises UnsetValueError if the value is Undefined.
             try:
                 model_val = getattr(model, attr)
@@ -327,10 +367,6 @@ class Syncable(Renderable):
             if not model.lookup(attr).property.matches(model_val, value):
                 attrs.append(attr)
 
-            # Do not apply model change that is in flight
-            if attr in self._events:
-                del self._events[attr]
-
         try:
             model.update(**msg)
         finally:
@@ -352,6 +388,11 @@ class Syncable(Renderable):
             model, _ = self._models.pop(ref, None)
             model._callbacks = {}
             model._event_callbacks = {}
+        if not self._models and self._watching_stylesheets:
+            self._watching_stylesheets.set()
+            if self._watching_stylesheets in state._watch_events:
+                state._watch_events.remove(self._watching_stylesheets)
+            self._watching_stylesheets = False
         comm, client_comm = self._comms.pop(ref, (None, None))
         if comm:
             try:
@@ -370,20 +411,50 @@ class Syncable(Renderable):
         changes = {event.name: event.new for event in events}
         return self._process_param_change(changes)
 
+    def _setup_autoreload(self):
+        from .config import config
+        paths = [sts for sts in self._stylesheets if isinstance(sts, pathlib.PurePath)]
+        if (self._watching_stylesheets or not (config.autoreload and paths and import_available('watchfiles'))):
+            return
+        self._watching_stylesheets = event = asyncio.Event()
+        state._watch_events.append(event)
+        state.execute(self._watch_stylesheets)
+
+    async def _watch_stylesheets(self):
+        import watchfiles
+        base_dir = pathlib.Path(inspect.getfile(type(self))).parent
+        paths = []
+        for sts in self._stylesheets:
+            if isinstance(sts, pathlib.PurePath):
+                if not sts.absolute().is_file():
+                    sts = base_dir / sts
+                if sts.is_file():
+                    paths.append(sts)
+        async for _ in watchfiles.awatch(*paths, stop_event=self._watching_stylesheets):
+            self.param.trigger('stylesheets')
+
     def _param_change(self, *events: param.parameterized.Event) -> None:
         named_events = {event.name: event for event in events}
+        applied = False
         for ref, (model, _) in self._models.copy().items():
             properties = self._update_properties(*events, doc=model.document)
             if not properties:
                 return
-            self._apply_update(named_events, properties, model, ref)
+            applied &= self._apply_update(named_events, properties, model, ref)
+            if ref not in state._views:
+                continue
+            doc = state._views[ref][2]
+            if applied and doc in self._in_process__events:
+                del self._in_process__events[doc]
 
     def _process_events(self, events: dict[str, Any]) -> None:
         self._log('received events %s', events)
         if any(e for e in events if e not in self._busy__ignore):
             with edit_readonly(state):
                 state._busy_counter += 1
-        params = self._process_property_change(dict(events))
+        if events:
+            self._in_process__events[state.curdoc] = events
+        params = self._process_property_change(events)
         try:
             with edit_readonly(self):
                 self_params = {k: v for k, v in params.items() if '.' not in k}
@@ -549,7 +620,8 @@ class Reactive(Syncable, Viewable):
 
     def __init__(self, refs=None, **params):
         for name, pobj in self.param.objects('existing').items():
-            if name not in self._param__private.explicit_no_refs:
+            if (name not in self._param__private.explicit_no_refs and
+                not isinstance(pobj, Child)):
                 pobj.allow_refs = True
         if refs is not None:
             self._refs = refs
@@ -655,7 +727,7 @@ class Reactive(Syncable, Viewable):
           Maps from a parameter in the source object to a callback.
         bidirectional: bool
           Whether to link source and target bi-directionally
-        **links: dict
+        links: dict
           Maps between parameters on this object to the parameters
           on the supplied object.
         """
@@ -709,7 +781,7 @@ class Reactive(Syncable, Viewable):
         self._links.append(link)
         return cb
 
-    def controls(self, parameters: list[str] = [], jslink: bool = True, **kwargs) -> 'Panel':
+    def controls(self, parameters: list[str] = [], jslink: bool = True, **kwargs) -> BasePanel:
         """
         Creates a set of widgets which allow manipulating the parameters
         on this instance. By default all parameters which support
@@ -779,7 +851,7 @@ class Reactive(Syncable, Viewable):
         ----------
         args: dict
           A mapping of objects to make available to the JS callback
-        **callbacks: dict
+        callbacks: dict
           A mapping between properties on the source model and the code
           to execute when that property changes
 
@@ -816,7 +888,7 @@ class Reactive(Syncable, Viewable):
           A mapping of objects to make available to the JS callback
         bidirectional: boolean
           Whether to link source and target bi-directionally
-        **links: dict
+        links: dict
           A mapping between properties on the source model and the
           target model property to link it to.
 
@@ -844,6 +916,34 @@ class Reactive(Syncable, Viewable):
         return Link(self, target, properties=links, code=code, args=args,
                     bidirectional=bidirectional)
 
+    def _send_event(self, Event: ModelEvent, **event_kwargs):
+        """
+        Send an event to the frontend
+
+        Arguments
+        ----------
+        Event: Bokeh.Event
+            The event to send to the frontend
+        event_kwargs: dict
+            Additional keyword arguments to pass to the event
+            This will create the following event:
+            Event(model=model, **event_kwargs)
+
+        """
+        for ref, (model, _) in self._models.copy().items():
+            if ref not in state._views or ref in state._fake_roots:
+                continue
+            event = Event(model=model, **event_kwargs)
+            _viewable, root, doc, comm = state._views[ref]
+            if comm or state._unblocked(doc) or not doc.session_context:
+                with unlocked():
+                    doc.callbacks.send_event(event)
+                if comm and 'embedded' not in root.tags:
+                    push(doc, comm)
+            else:
+                cb = partial(doc.callbacks.send_event, event)
+                doc.add_next_tick_callback(cb)
+
 
 TData = Union['pd.DataFrame', 'DataDict']
 
@@ -932,7 +1032,7 @@ class SyncableData(Reactive):
         self._processed, self._data = self._get_data()
         msg = {'data': self._data}
         named_events = {event.name: event for event in events}
-        for ref, (m, _) in self._models.items():
+        for ref, (m, _) in self._models.copy().items():
             self._apply_update(named_events, msg, m.source, ref)
 
     @updating
@@ -942,7 +1042,7 @@ class SyncableData(Reactive):
         indices = self.selection if indices is None else indices
         msg = {'indices': indices}
         named_events = {event.name: event for event in events}
-        for ref, (m, _) in self._models.items():
+        for ref, (m, _) in self._models.copy().items():
             self._apply_update(named_events, msg, m.source.selected, ref)
 
     def _apply_stream(self, ref: str, model: Model, stream: 'DataDict', rollover: Optional[int]) -> None:
@@ -955,7 +1055,7 @@ class SyncableData(Reactive):
     @updating
     def _stream(self, stream: 'DataDict', rollover: Optional[int] = None) -> None:
         self._processed, _ = self._get_data()
-        for ref, (m, _) in self._models.items():
+        for ref, (m, _) in self._models.copy().items():
             if ref not in state._views or ref in state._fake_roots:
                 continue
             viewable, root, doc, comm = state._views[ref]
@@ -977,7 +1077,7 @@ class SyncableData(Reactive):
 
     @updating
     def _patch(self, patch: 'Patches') -> None:
-        for ref, (m, _) in self._models.items():
+        for ref, (m, _) in self._models.copy().items():
             if ref not in state._views or ref in state._fake_roots:
                 continue
             viewable, root, doc, comm = state._views[ref]
@@ -1213,9 +1313,6 @@ class ReactiveData(SyncableData):
 
     __abstract = True
 
-    def __init__(self, **params):
-        super().__init__(**params)
-
     def _update_selection(self, indices: list[int]) -> None:
         self.selection = indices
 
@@ -1227,21 +1324,14 @@ class ReactiveData(SyncableData):
         if dtype.kind == 'M':
             if values.dtype.kind in 'if':
                 if getattr(dtype, 'tz', None):
-                    # dtype has a timezone
-                    if dtype.tz == dt.timezone.utc:
-                        # Milliseconds to nanoseconds, to datetime64.
-                        converted = (values * 1e6).astype('datetime64[ns]')
-                    else:
-                        import pandas as pd
+                    import pandas as pd
 
-                        # Using pandas to convert from milliseconds
-                        # timezone-aware, to UTC nanoseconds, to datetime64.
-                        converted = (
-                            pd.Series(pd.to_datetime(values, unit="ms"))
-                            .dt.tz_localize(dtype.tz)
-                            .dt.tz_convert('utc')
-                            .dt.tz_localize(None)
-                        )
+                    # Using pandas to convert from milliseconds
+                    # timezone-aware, to UTC nanoseconds, to datetime64.
+                    converted = (
+                        pd.Series(pd.to_datetime(values, unit="ms"))
+                        .dt.tz_localize(dtype.tz)
+                    )
                 else:
                     # Timestamps converted from milliseconds to nanoseconds,
                     # to datetime.
@@ -1339,18 +1429,20 @@ class ReactiveData(SyncableData):
         super(ReactiveData, self)._process_events(events)
 
 
+class ReactiveMetaBase(ParameterizedMetaclass):
 
-class ReactiveHTMLMetaclass(ParameterizedMetaclass):
+    _loaded_extensions: ClassVar[set[str]] = set()
+
+    _name_counter: ClassVar[Counter] = Counter()
+
+
+class ReactiveHTMLMetaclass(ReactiveMetaBase):
     """
     Parses the ReactiveHTML._template of the class and initializes
     variables, callbacks and the data model to sync the parameters and
     HTML attributes.
     """
 
-    _loaded_extensions: ClassVar[set[str]] = set()
-
-    _name_counter: ClassVar[Counter] = Counter()
-
     _script_regex: ClassVar[str] = r"script\([\"|'](.*)[\"|']\)"
 
     def __init__(mcs, name: str, bases: tuple[type, ...], dict_: Mapping[str, Any]):
@@ -1447,13 +1539,64 @@ class ReactiveHTMLMetaclass(ParameterizedMetaclass):
         # Create model with unique name
         ReactiveHTMLMetaclass._name_counter[name] += 1
         model_name = f'{name}{ReactiveHTMLMetaclass._name_counter[name]}'
+
         mcs._data_model = construct_data_model(
             mcs, name=model_name, ignore=ignored, types=types
         )
 
 
+class ReactiveCustomBase(Reactive):
 
-class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
+    _extension_name: ClassVar[Optional[str]] = None
+
+    __css__: ClassVar[Optional[list[str]]] = None
+    __javascript__: ClassVar[Optional[list[str]]] = None
+    __javascript_modules__: ClassVar[Optional[list[str]]] = None
+
+    @classmethod
+    def _loaded(cls) -> bool:
+        """
+        Whether the component has been loaded.
+        """
+        return (
+            cls._extension_name is None or
+            (cls._extension_name in ReactiveMetaBase._loaded_extensions and
+             (state._extensions is None or (cls._extension_name in state._extensions)))
+        )
+
+    def _process_param_change(self, params):
+        props = super()._process_param_change(params)
+        if 'stylesheets' in params:
+            css = getattr(self, '__css__', []) or []
+            if state.rel_path:
+                css = [
+                    ss if ss.startswith('http') else f'{state.rel_path}/{ss}'
+                    for ss in css
+                ]
+            props['stylesheets'] = [
+                ImportedStyleSheet(url=ss) for ss in css
+            ] + props['stylesheets']
+        return props
+
+    def _set_on_model(self, msg: Mapping[str, Any], root: Model, model: Model) -> None:
+        if not msg:
+            return
+        old = self._changing.get(root.ref['id'], [])
+        self._changing[root.ref['id']] = [
+            attr for attr, value in msg.items()
+            if not model.lookup(attr).property.matches(getattr(model, attr), value)
+        ]
+        try:
+            model.update(**msg)
+        finally:
+            if old:
+                self._changing[root.ref['id']] = old
+            else:
+                del self._changing[root.ref['id']]
+
+
+
+class ReactiveHTML(ReactiveCustomBase, metaclass=ReactiveHTMLMetaclass):
     """
     The ReactiveHTML class enables you to create custom Panel components using HTML, CSS and/ or
     Javascript and without the complexities of Javascript build tools.
@@ -1594,8 +1737,6 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
 
     _dom_events: ClassVar[Mapping[str, list[str]]] = {}
 
-    _extension_name: ClassVar[Optional[str]] = None
-
     _template: ClassVar[str] = ""
 
     _scripts: ClassVar[Mapping[str, str | list[str]]] = {}
@@ -1604,14 +1745,11 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         r'data\.([^[^\d\W]\w*)[ ]*[\+,\-,\*,\\,%,\*\*,<<,>>,>>>,&,\^,|,\&\&,\|\|,\?\?]*='
     )
 
-    __css__: ClassVar[Optional[list[str]]] = None
-    __javascript__: ClassVar[Optional[list[str]]] = None
-    __javascript_modules__: ClassVar[Optional[list[str]]] = None
-
     __abstract = True
 
     def __init__(self, **params):
         from .pane import panel
+        cls = type(self)
         for children_param in self._parser.children.values():
             mode = self._child_config.get(children_param, 'model')
             if children_param not in params or mode != 'model':
@@ -1632,23 +1770,28 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
                     children[key] = panel(pane)
                 params[children_param] = children
             else:
-                params[children_param] = panel(child_value)
+                params[children_param] = children = panel(child_value)
+            child_param = cls.param[children_param]
+            try:
+                if children_param not in self._param__private.explicit_no_refs:
+                    children = resolve_value(children)
+                if not ((isinstance(child_param, Children) and isinstance(children, list)) or
+                        (isinstance(child_param, Child))):
+                    child_param._validate(children)
+            except Exception as e:
+                raise RuntimeError(
+                    f"{cls.__name__}._template declares {children_param!r} "
+                    "parameter as a child, however the parameter is of type "
+                    f"{type(child_param).__name__}. Either ensure that the parameter "
+                    "can accept a Panel component, use literal template using the "
+                    "Jinja2 templating syntax (i.e. {{ <param> }}) or declare the "
+                    "child as a literal in the _child_config."
+                ) from e
         super().__init__(**params)
         self._attrs = {}
         self._panes = {}
         self._event_callbacks = defaultdict(lambda: defaultdict(list))
 
-    @classmethod
-    def _loaded(cls) -> bool:
-        """
-        Whether the component has been loaded.
-        """
-        return (
-            cls._extension_name is None or
-            (cls._extension_name in ReactiveHTMLMetaclass._loaded_extensions and
-             (state._extensions is None or (cls._extension_name in state._extensions)))
-        )
-
     def _cleanup(self, root: Model | None = None) -> None:
         for _child, panes in self._panes.items():
             for pane in panes:
@@ -1803,8 +1946,6 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         return self._process_children(doc, root, model, comm, new_models)
 
     def _get_template(self) -> tuple[str, list[str], Mapping[str, list[tuple[str, list[str], str]]]]:
-        import jinja2
-
         # Replace loop variables with indexed child parameter e.g.:
         #   {% for obj in objects %}
         #     ${obj}
@@ -1817,11 +1958,11 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         for parent_var, obj in self._parser.loop_map.items():
             for var in self._parser.loop_var_map[parent_var]:
                 template_string = template_string.replace(
-                    '${%s}' % var, '${%s[{{ loop.index0 }}]}' % obj)
+                    '${%s}' % var, '${%s[{{ loop.index0 }}]}' % obj)  # noqa: UP031
 
         # Add index to templated loop node ids
         for dom_node, _ in self._parser.looped:
-            replacement = 'id="%s-{{ loop.index0 }}"' % dom_node
+            replacement = 'id="%s-{{ loop.index0 }}"' % dom_node  # noqa: UP031
             if f'id="{dom_node}"' in template_string:
                 template_string = template_string.replace(
                     f'id="{dom_node}"', replacement)
@@ -1875,7 +2016,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
                 for i, _ in enumerate(getattr(self, child_name)):
                     html = html.replace('${%s[%d]}' % (child_name, i), '')
             else:
-                html = html.replace('${%s}' % child_name, '')
+                html = html.replace(f'${{{child_name}}}', '')
         return html, parser.nodes, p_attrs
 
     @property
@@ -1928,18 +2069,21 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
                 f'\n\npn.extension(\'{self._extension_name}\')\n'
             )
         if self._extension_name:
-            ReactiveHTMLMetaclass._loaded_extensions.add(self._extension_name)
+            ReactiveMetaBase._loaded_extensions.add(self._extension_name)
 
         if not root:
             root = model
 
         ref = root.ref['id']
         data_model: DataModel = model.data # type: ignore
+        for v in data_model.properties_with_values():
+            if isinstance(v, DataModel):
+                v.tags.append(f"__ref:{ref}")
         self._patch_datamodel_ref(data_model.properties_with_values(), ref)
         model.update(children=self._get_children(doc, root, model, comm))
         self._register_events('dom_event', model=model, doc=doc, comm=comm)
         self._link_props(data_model, self._linked_properties, doc, root, comm)
-        self._models[ref] = (model, parent)
+        self._models[root.ref['id']] = (model, parent)
         return model
 
     def _process_event(self, event: 'Event') -> None:
@@ -1966,23 +2110,9 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
             cb(event)
 
     def _set_on_model(self, msg: Mapping[str, Any], root: Model, model: Model) -> None:
-        if not msg:
-            return
-        ref = root.ref['id']
-        old = self._changing.get(ref, [])
-        self._changing[ref] = [
-            attr for attr, value in msg.items()
-            if not model.lookup(attr).property.matches(getattr(model, attr), value)
-        ]
-        try:
-            model.update(**msg)
-        finally:
-            if old:
-                self._changing[ref] = old
-            else:
-                del self._changing[ref]
+        super()._set_on_model(msg, root, model)
         if isinstance(model, DataModel):
-            self._patch_datamodel_ref(model.properties_with_values(), ref)
+            self._patch_datamodel_ref(model.properties_with_values(), root.ref['id'])
 
     def _update_model(
         self, events: dict[str, param.parameterized.Event], msg: dict[str, Any],
@@ -2049,7 +2179,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
                              f"nodes include: {self._parser.nodes}.")
         self._event_callbacks[node][event].append(callback)
         events = self._get_events()
-        for ref, (model, _) in self._models.items():
+        for ref, (model, _) in self._models.copy().items():
             self._apply_update({}, {'events': events}, model, ref)
 
 __all__ = (
