@@ -8,7 +8,7 @@ import ast
 import json
 
 from base64 import b64decode
-from datetime import date, datetime
+from datetime import date, datetime, time as dt_time
 from typing import (
     TYPE_CHECKING, Any, ClassVar, Iterable, Mapping, Optional,
 )
@@ -22,15 +22,20 @@ from bokeh.models.widgets import (
     DatePicker as _BkDatePicker, DateRangePicker as _BkDateRangePicker,
     Div as _BkDiv, FileInput as _BkFileInput, NumericInput as _BkNumericInput,
     PasswordInput as _BkPasswordInput, Spinner as _BkSpinner,
-    Switch as _BkSwitch, TextInput as _BkTextInput,
+    Switch as _BkSwitch,
 )
+from bokeh.models.widgets.inputs import ClearInput
+from pyviz_comms import JupyterComm
 
 from ..config import config
 from ..layout import Column, Panel
 from ..models import (
     DatetimePicker as _bkDatetimePicker, TextAreaInput as _bkTextAreaInput,
+    TextInput as _BkTextInput, TimePicker as _BkTimePicker,
+)
+from ..util import (
+    escape, lazy_load, param_reprs, try_datetime64_to_datetime,
 )
-from ..util import param_reprs, try_datetime64_to_datetime
 from .base import CompositeWidget, Widget
 
 if TYPE_CHECKING:
@@ -38,19 +43,11 @@ if TYPE_CHECKING:
     from bokeh.model import Model
     from pyviz_comms import Comm
 
+    from ..models.file_dropper import DeleteEvent, UploadEvent
     from ..viewable import Viewable
 
 
-class TextInput(Widget):
-    """
-    The `TextInput` widget allows entering any string using a text input box.
-
-    Reference: https://panel.holoviz.org/reference/widgets/TextInput.html
-
-    :Example:
-
-    >>> TextInput(name='Name', placeholder='Enter your name here ...')
-    """
+class _TextInputBase(Widget):
 
     description = param.String(default=None, doc="""
         An HTML string describing the function of this component.""")
@@ -71,8 +68,6 @@ class TextInput(Widget):
       Width of this component. If sizing_mode is set to stretch
       or scale mode this will merely be used as a suggestion.""")
 
-    _widget_type: ClassVar[type[Model]] = _BkTextInput
-
     @classmethod
     def from_param(cls, parameter: param.Parameter, onkeyup=False, **params) -> Viewable:
         """
@@ -96,7 +91,41 @@ class TextInput(Widget):
         return super().from_param(parameter, **params)
 
 
-class PasswordInput(TextInput):
+class TextInput(_TextInputBase):
+
+    """
+    The `TextInput` widget allows entering any string using a text input box.
+
+    Reference: https://panel.holoviz.org/reference/widgets/TextInput.html
+
+    :Example:
+
+    >>> TextInput(name='Name', placeholder='Enter your name here ...')
+    """
+
+    enter_pressed = param.Event(doc="""
+        Event when the enter key has been pressed.""")
+
+    _widget_type: ClassVar[type[Model]] = _BkTextInput
+
+    _rename = {'enter_pressed': None}
+
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
+        model = super()._get_model(doc, root, parent, comm)
+        self._register_events('enter-pressed', model=model, doc=doc, comm=comm)
+        return model
+
+    def _process_event(self, event) -> None:
+        if event.event_name == 'enter-pressed':
+            self.value = event.value_input
+            self.value_input = event.value_input
+            self.enter_pressed = True
+
+
+class PasswordInput(_TextInputBase):
     """
     The `PasswordInput` allows entering any string using an obfuscated text
     input box.
@@ -113,7 +142,7 @@ class PasswordInput(TextInput):
     _widget_type: ClassVar[type[Model]] = _BkPasswordInput
 
 
-class TextAreaInput(TextInput):
+class TextAreaInput(_TextInputBase):
     """
     The `TextAreaInput` allows entering any multiline string using a text input
     box.
@@ -169,20 +198,43 @@ class FileInput(Widget):
     >>> FileInput(accept='.png,.jpeg', multiple=True)
     """
 
-    accept = param.String(default=None)
+    accept = param.String(default=None, doc="""
+        A comma separated string of all extension types that should
+        be supported.""")
 
     description = param.String(default=None, doc="""
-        An HTML string describing the function of this component.""")
+        An HTML string describing the function of this component
+        rendered as a tooltip icon.""")
+
+    directory = param.Boolean(default=False, doc="""
+        Whether to allow selection of directories instead of files.
+        The filename will be relative paths to the uploaded directory.
+
+        .. note::
+            When a directory is uploaded it will give add a confirmation pop up.
+            The confirmation pop up cannot be disabled, as this is a security feature
+            in the browser.
+
+        .. note::
+            The `accept` parameter only works with file extension.
+            When using `accept` with `directory`, the number of files
+            reported will be the total amount of files, not the filtered.""")
 
     filename = param.ClassSelector(
-        default=None, class_=(str, list), is_instance=True)
+        default=None, class_=(str, list), is_instance=True, doc="""
+        Name of the uploaded file(s).""")
 
     mime_type = param.ClassSelector(
-        default=None, class_=(str, list), is_instance=True)
+        default=None, class_=(str, list), is_instance=True, doc="""
+        Mimetype of the uploaded file(s).""")
 
-    multiple = param.Boolean(default=False)
+    multiple = param.Boolean(default=False, doc="""
+        Whether to allow uploading multiple files. If enabled value
+        parameter will return a list.""")
 
-    value = param.Parameter(default=None)
+    value = param.Parameter(default=None, doc="""
+        The uploaded file(s) stored as a single bytes object if
+        multiple is False or a list of bytes otherwise.""")
 
     _rename: ClassVar[Mapping[str, str | None]] = {
         'filename': None, 'name': None
@@ -211,9 +263,13 @@ class FileInput(Widget):
         msg = super()._process_property_change(msg)
         if 'value' in msg:
             if isinstance(msg['value'], str):
-                msg['value'] = b64decode(msg['value'])
+                msg['value'] = b64decode(msg['value']) if msg['value'] else None
             else:
                 msg['value'] = [b64decode(content) for content in msg['value']]
+        if 'filename' in msg and len(msg['filename']) == 0:
+            msg['filename'] = None
+        if 'mime_type' in msg and len(msg['mime_type']) == 0:
+            msg['mime_type'] = None
         return msg
 
     def save(self, filename):
@@ -250,6 +306,118 @@ class FileInput(Widget):
             else:
                 fn.write(val)
 
+    def clear(self):
+        """
+        Clear the file(s) in the FileInput widget
+        """
+        self._send_event(ClearInput)
+
+
+class FileDropper(Widget):
+    """
+    The `FileDropper` allows the user to upload one or more files to the server.
+
+    It is similar to the `FileInput` widget but additionally adds support
+    for chunked uploads, making it possible to upload large files. The
+    UI also supports previews for image files. Unlike `FileInput` the
+    uploaded files are stored as dictionary of bytes object indexed
+    by the filename.
+
+    Reference: https://panel.holoviz.org/reference/widgets/FileDropper.html
+
+    :Example:
+
+    >>> FileDropper(accepted_filetypes=['image/*'], multiple=True)
+    """
+
+    accepted_filetypes = param.List(default=[], doc="""
+        List of accepted file types. Can be mime types, file extensions
+        or wild cards.For instance ['image/*'] will accept all images.
+        ['.png', 'image/jpeg'] will only accepts PNGs and JPEGs.""")
+
+    chunk_size = param.Integer(default=10_000_000, doc="""
+        Size in bytes per chunk transferred across the WebSocket.""")
+
+    layout = param.Selector(
+        default=None, objects=["circle", "compact", "integrated"], doc="""
+        Compact mode will remove padding, integrated mode is used to render
+        FilePond as part of a bigger element. Circle mode adjusts the item
+        position offsets so buttons and progress indicators don't fall outside
+        of the circular shape.""")
+
+    max_file_size = param.String(default=None, doc="""
+        Maximum size of a file as a string with units given in KB or MB,
+        e.g. 5MB or 750KB.""")
+
+    max_files = param.Integer(default=None, doc="""
+        Maximum number of files that can be uploaded if multiple=True.""")
+
+    max_total_file_size = param.String(default=None, doc="""
+        Maximum size of all uploaded files, as a string with units given
+        in KB or MB, e.g. 5MB or 750KB.""")
+
+    mime_type = param.Dict(default={}, doc="""
+        A dictionary containing the mimetypes for each of the uploaded
+        files indexed by their filename.""")
+
+    multiple = param.Boolean(default=False, doc="""
+        Whether to allow uploading multiple files.""")
+
+    value = param.Dict(default={}, doc="""
+        A dictionary containing the uploaded file(s) as bytes or string
+        objects indexed by the filename. Files that have a text/* mimetype
+        will automatically be decoded as utf-8.""")
+
+    width = param.Integer(default=300, allow_None=True, doc="""
+      Width of this component. If sizing_mode is set to stretch
+      or scale mode this will merely be used as a suggestion.""")
+
+    _rename = {'value': None}
+
+    def __init__(self, **params):
+        super().__init__(**params)
+        self._file_buffer = {}
+
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
+        self._widget_type = lazy_load(
+            'panel.models.file_dropper', 'FileDropper', isinstance(comm, JupyterComm), root,
+            ext='filedropper'
+        )
+        model = super()._get_model(doc, root, parent, comm)
+        self._register_events('delete_event', 'upload_event', model=model, doc=doc, comm=comm)
+        return model
+
+    def _process_event(self, event: DeleteEvent | UploadEvent):
+        data = event.data
+        name = data['name']
+        if event.event_name == 'delete_event':
+            if name in self.mime_type:
+                del self.mime_type[name]
+            if name in self.value:
+                del self.value[name]
+            self.param.trigger('mime_type', 'value')
+            return
+
+        if data['chunk'] == 1:
+            self._file_buffer[name] = []
+        self._file_buffer[name].append(data['data'])
+        if data['chunk'] != data['total_chunks']:
+            return
+
+        buffers = self._file_buffer.pop(name)
+        file_buffer = b''.join(buffers)
+        if data['type'].startswith('text/'):
+            try:
+                file_buffer = file_buffer.decode('utf-8')
+            except UnicodeDecodeError:
+                pass
+        self.value[name] = file_buffer
+        self.mime_type[name] = data['type']
+        self.param.trigger('mime_type', 'value')
+
 
 class StaticText(Widget):
     """
@@ -264,7 +432,7 @@ class StaticText(Widget):
     """
 
     value = param.Parameter(default=None, doc="""
-        The current value""")
+        The current value to be displayed.""")
 
     _format: ClassVar[str] = '<b>{title}</b>: {value}'
 
@@ -280,10 +448,22 @@ class StaticText(Widget):
 
     _widget_type: ClassVar[type[Model]] = _BkDiv
 
+    @property
+    def _linked_properties(self) -> tuple[str]:
+        return ()
+
+    def _init_params(self) -> dict[str, Any]:
+        return {
+            k: v for k, v in self.param.values().items()
+            if k in self._synced_params and (v is not None or k == 'value')
+        }
+
     def _process_param_change(self, msg):
         msg = super()._process_param_change(msg)
         if 'text' in msg:
-            text = str(msg.pop('text'))
+            text = msg.pop('text')
+            if not isinstance(text, str):
+                text = escape("" if text is None else str(text))
             partial = self._format.replace('{value}', '').format(title=self.name)
             if self.name:
                 text = self._format.format(title=self.name, value=text.replace(partial, ''))
@@ -427,8 +607,11 @@ class DateRangePicker(Widget):
 
     def _process_param_change(self, msg):
         msg = super()._process_param_change(msg)
-        if 'value' in msg:
-            msg['value'] = tuple(self._convert_date_to_string(v) for v in msg['value'])
+        if 'value' in msg and msg['value'] is not None:
+            msg['value'] = tuple(
+                v if v is None else self._convert_date_to_string(v)
+                for v in msg['value']
+            )
         if 'min_date' in msg:
             msg['min_date'] = self._convert_date_to_string(msg['min_date'])
         if 'max_date' in msg:
@@ -628,6 +811,77 @@ class DatetimeRangePicker(_DatetimePickerBase):
         return value
 
 
+class _TimeCommon(Widget):
+
+    hour_increment = param.Integer(default=1, doc="""
+    Defines the granularity of hour value increments in the UI.
+    """)
+
+    minute_increment = param.Integer(default=1, doc="""
+    Defines the granularity of minute value increments in the UI.
+    """)
+
+    second_increment = param.Integer(default=1, doc="""
+    Defines the granularity of second value increments in the UI.
+    """)
+
+    seconds = param.Boolean(default=False, doc="""
+    Allows to select seconds. By default only hours and minutes are
+    selectable, and AM/PM depending on the `clock` option.
+    """)
+
+    clock = param.String(default='12h', doc="""
+        Whether to use 12 hour or 24 hour clock.""")
+
+    __abstract = True
+
+
+class TimePicker(_TimeCommon):
+    """
+    The `TimePicker` allows selecting a `time` value using a text box
+    and a time-picking utility.
+
+    Reference: https://panel.holoviz.org/reference/widgets/TimePicker.html
+
+    :Example:
+
+    >>> TimePicker(
+    ...     value="12:59:31", start="09:00:00", end="18:00:00", name="Time"
+    ... )
+    """
+
+    value = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
+        The current value""")
+
+    start = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
+        Inclusive lower bound of the allowed time selection""")
+
+    end = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
+        Inclusive upper bound of the allowed time selection""")
+
+    format = param.String(default='H:i', doc="""
+        Formatting specification for the display of the picked date.
+
+        +---+------------------------------------+------------+
+        | H | Hours (24 hours)                   | 00 to 23   |
+        | h | Hours                              | 1 to 12    |
+        | G | Hours, 2 digits with leading zeros | 1 to 12    |
+        | i | Minutes                            | 00 to 59   |
+        | S | Seconds, 2 digits                  | 00 to 59   |
+        | s | Seconds                            | 0, 1 to 59 |
+        | K | AM/PM                              | AM or PM   |
+        +---+------------------------------------+------------+
+
+        See also https://flatpickr.js.org/formatting/#date-formatting-tokens.
+    """)
+
+    _rename: ClassVar[Mapping[str, str | None]] = {
+        'start': 'min_time', 'end': 'max_time', 'format': 'time_format'
+    }
+
+    _widget_type: ClassVar[type[Model]] = _BkTimePicker
+
+
 class ColorPicker(Widget):
     """
     The `ColorPicker` widget allows selecting a hexadecimal RGB color value
@@ -777,6 +1031,11 @@ class _SpinnerBase(_NumericInputBase):
                 msg["value"] = msg["value_throttled"]
         return super()._process_property_change(msg)
 
+    def _process_events(self, events: dict[str, Any]) -> None:
+        if config.throttled:
+            events.pop("value", None)
+        super()._process_events(events)
+
 
 class IntInput(_SpinnerBase, _IntInputBase):
     """
